class FinancialSymbolId:
    __delimiter = '/'

    @classmethod
    def parse(cls, fin_symbol_fullname: str):
        namespace, name = fin_symbol_fullname.split(FinancialSymbolId.__delimiter)
        return FinancialSymbolId(namespace, name)

    def __init__(self, namespace: str, name: str):
        self.namespace = namespace
        self.name = name

    def format(self) -> str:
        return self.namespace + FinancialSymbolId.__delimiter + self.name

    def __repr__(self):
        return self.format()



from functools import lru_cache
from typing import Optional, Callable

import pandas as pd
from serum import singleton

from .base_classes import FinancialSymbolsSource
from .._settings import data_url
from ..common.enums import Currency, SecurityType, Period
from ..common.financial_symbol import FinancialSymbol
from ..common.financial_symbol_id import FinancialSymbolId
from ..common.financial_symbol_info import FinancialSymbolInfo


@singleton
class MutualFundsRuSource(FinancialSymbolsSource):
    def __init__(self):
        super().__init__(namespace='mut_ru')
        self.url_base = data_url + 'mut_ru/'

        self.index = pd.read_csv(self.url_base + '__index.csv', sep='\t', index_col='name',
                                 parse_dates=['date_start', 'date_end'])
        self.index['date_start'] = self.index['date_start'].dt.to_period(freq='D')
        self.index['date_end'] = self.index['date_end'].dt.to_period(freq='D')

    @lru_cache(maxsize=512)
    def __extract_values(self, row_id: str) -> Callable[[pd.Period, pd.Period], pd.DataFrame]:
        def func(start_period: pd.Period, end_period: pd.Period) -> pd.DataFrame:
            url = '{}{}.csv'.format(self.url_base, row_id)
            df = pd.read_csv(url, sep='\t', parse_dates=['date'], usecols=['date', 'adjusted_close'])
            df.rename(columns={'adjusted_close': 'close'}, inplace=True)
            df['period'] = df['date'].dt.to_period('M')
            df_new = df[(start_period <= df['period']) & (df['period'] <= end_period)].copy()
            return df_new

        return func

    def fetch_financial_symbol(self, name: str) -> Optional[FinancialSymbol]:
        if name not in self.index.index:
            return None
        row = self.index.loc[name]
        symbol = FinancialSymbol(identifier=FinancialSymbolId(namespace=self.namespace, name=name),
                                 values=self.__extract_values(name),
                                 short_name=row['short_name'],
                                 start_period=row['date_start'],
                                 end_period=row['date_end'],
                                 currency=Currency.RUB,
                                 security_type=SecurityType.MUT,
                                 period=Period.DAY,
                                 adjusted_close=True)
        return symbol

    def get_all_infos(self):
        infos = []
        for idx, row in self.index.iterrows():
            fin_sym_info = FinancialSymbolInfo(
                fin_sym_id=FinancialSymbolId(self.namespace, str(idx)),
                short_name=row['short_name']
            )
            infos.append(fin_sym_info)
        return infos



from typing import Optional, List, Dict

from serum import singleton, inject
import pandas as pd
from itertools import groupby

from .._sources.base_classes import FinancialSymbolsSource
from .._sources.single_financial_symbol_source import CbrCurrenciesSource
from .all_sources import SymbolSources
from .._settings import data_url
from ..common.financial_symbol import FinancialSymbol
from ..common.financial_symbol_id import FinancialSymbolId
from ..common.enums import Currency


@singleton
class FinancialSymbolsRegistry:

    @inject
    def __init__(self, symbol_sources: SymbolSources):
        def symbol_source_key(x: FinancialSymbolsSource) -> str:
            return x.namespace

        symbol_sources_list: List[FinancialSymbolsSource] = \
            sorted(symbol_sources.sources, key=symbol_source_key)
        self.symbol_sources: Dict[str, List[FinancialSymbolsSource]] = {}
        for k, v in groupby(symbol_sources_list, key=symbol_source_key):
            self.symbol_sources.update({k: list(v)})

    def namespaces(self):
        return list(self.symbol_sources.keys())

    def get_all_infos(self, namespace):
        if namespace is None:
            return None
        return [name
                for sym_source in self.symbol_sources.get(namespace)
                for name in sym_source.get_all_infos()]

    def get(self, financial_symbol_id: FinancialSymbolId) -> Optional[FinancialSymbol]:
        symbol_sources_list: Optional[List[FinancialSymbolsSource]] = \
            self.symbol_sources.get(financial_symbol_id.namespace)

        if symbol_sources_list is None:
            return None

        result: List[FinancialSymbol] = []
        for symbol_source in symbol_sources_list:
            fin_symbol = symbol_source.fetch_financial_symbol(financial_symbol_id.name)
            if fin_symbol is not None:
                result.append(fin_symbol)
        result_count = len(result)
        if result_count == 1:
            return result[0]
        elif result_count == 0:
            return None
        else:
            raise Exception('Something went wrong. {} names are found for {}. '
                            'Please, submit the issue'
                            .format(result_count, financial_symbol_id.format()))


@singleton
@inject
class CurrencySymbolsRegistry:
    cbr_currencies_source: CbrCurrenciesSource

    def __init__(self):
        self.url_base = data_url + 'currency/'
        currency_index = pd.read_csv('{}__index.csv'.format(self.url_base),
                                     sep='\t', parse_dates=['date_start', 'date_end'])
        self.__f_currency_data = {}
        for supported_currency_pair in currency_index['name']:
            url = '{}{}.csv'.format(self.url_base, supported_currency_pair)
            df = pd.read_csv(url, sep='\t', parse_dates=['date'])
            df['period'] = df['date'].dt.to_period('M')
            vals_lastdate_indices = df.groupby(['period'])['date'].transform(max) == df['date']
            df = df[vals_lastdate_indices].copy()
            df['close'] = df['close'] * df['nominal']
            del df['date'], df['nominal']
            supported_currency_pair = tuple(str.split(supported_currency_pair, '-'))
            self.__f_currency_data.update({supported_currency_pair: df})

    def __currency_data(self, currency_pair):
        return self.__f_currency_data[currency_pair].copy()

    def convert(self, currency_from: Currency, currency_to: Currency,
                start_period: pd.Period, end_period: pd.Period):
        if currency_to == currency_from:
            currency_min_date = self.cbr_currencies_source._currency_min_date[currency_from.name]
            p_range = pd.period_range(start=str(currency_min_date),
                                      end=str(end_period + 1),
                                      freq='M')
            df = pd.DataFrame.from_dict({'period': p_range, 'close': 1.})
        elif currency_to == Currency.RUB:
            df = self.__currency_data((currency_from.name, currency_to.name))
        elif currency_from == Currency.RUB:
            df = self.convert(currency_to, currency_from, start_period, end_period)
            df['close'] = 1.0 / df['close']
        else:
            df = self.convert(currency_from, Currency.RUB, start_period, end_period)
            df_to = self.convert(Currency.RUB, currency_to, start_period, end_period)
            df = df.merge(df_to, on='period', suffixes=('', '_to'))
            df['close'] = df['close'] * df['close_to']
            del df['close_to']

        df = df[(start_period <= df['period']) & (df['period'] <= end_period)].copy()
        df.sort_values(by='period', ascending=True, inplace=True)
        return df



from pprint import pformat


class FinancialSymbolInfo:
    def __init__(self, fin_sym_id, short_name):
        self.fin_sym_id = fin_sym_id
        self.short_name = short_name

    def __repr__(self) -> str:
        return pformat(vars(self))



from abc import ABCMeta, abstractmethod
from typing import List

from serum import singleton, inject

from .._sources.base_classes import FinancialSymbolsSource
from .micex_stocks_source import MicexStocksSource
from .mutru_funds_source import MutualFundsRuSource
from .quandl_source import QuandlSource
from .moex_indexes_source import MoexIndexesSource
from .single_financial_symbol_source import CbrCurrenciesSource, CbrTopRatesSource
from .okama_source import OkamaSource
from .inflation_source import InflationSource
from .yahoo_indexes_source import YahooIndexesSource


@singleton
class SymbolSources(metaclass=ABCMeta):
    @property
    @abstractmethod
    def sources(self) -> List[FinancialSymbolsSource]:
        raise NotImplementedError()


@inject
class AllSymbolSources(SymbolSources):
    cbr_currencies_source: CbrCurrenciesSource
    cbr_top_rates_source: CbrTopRatesSource
    inflation_source: InflationSource
    moex_indexes_source: MoexIndexesSource
    micex_stocks_source: MicexStocksSource
    mut_ru_source: MutualFundsRuSource
    quandl_source: QuandlSource
    okama_source: OkamaSource
    yahoo_indexes_source: YahooIndexesSource

    @property
    def sources(self):
        return [
            self.cbr_currencies_source,
            self.cbr_top_rates_source,
            self.inflation_source,
            self.micex_stocks_source,
            self.moex_indexes_source,
            self.mut_ru_source,
            self.quandl_source,
            self.okama_source,
            self.yahoo_indexes_source,
        ]



from typing import List, Dict, Union, Optional

import numpy as np
import pandas as pd
from contracts import contract
from serum import inject, singleton

from ._portfolio.currency import PortfolioCurrencyFactory
from ._portfolio.portfolio import Portfolio, PortfolioAsset
from ._search import _Search
from ._sources.registries import FinancialSymbolsRegistry
from .common.enums import Currency, SecurityType
from .common.financial_symbol import FinancialSymbol
from .common.financial_symbol_id import FinancialSymbolId


@singleton
class Yapo:

    @inject
    def __init__(self,
                 fin_syms_registry: FinancialSymbolsRegistry,
                 portfolio_currency_factory: PortfolioCurrencyFactory):
        self.__search = _Search()
        self.fin_syms_registry = fin_syms_registry
        self.portfolio_currency_factory = portfolio_currency_factory
        self.__period_lowest = '1900-1'
        self.__period_highest = lambda: str(pd.Period.now(freq='M'))

    def information(self, **kwargs) -> Union[Optional[FinancialSymbol], List[Optional[FinancialSymbol]]]:
        """
        Fetches financial symbol information based on internal ID.
        The information includes ISIN, short and long
        names, exchange, currency, etc.

        :param kwargs:
            either `name` or `names` should be defined

            name (str): name of a financial symbol

            names (List[str]): names of financial symbols

        :returns: financial symbol information
        """
        if 'name' in kwargs:
            name = kwargs['name']
            financial_symbol_id = FinancialSymbolId.parse(name)
            finsym_info = self.fin_syms_registry.get(financial_symbol_id)
            return finsym_info
        elif 'names' in kwargs:
            names = kwargs['names']
            finsym_infos: List[Optional[FinancialSymbol]] = []
            for name in names:
                finsym_info1 = self.information(name=name)
                if not (finsym_info1 is None or isinstance(finsym_info1, FinancialSymbol)):
                    raise ValueError('Unexpected type of financial symbol information')
                finsym_infos.append(finsym_info1)
            return finsym_infos
        else:
            raise Exception('Unexpected state of kwargs')

    def portfolio_asset(self,
                        currency: str = None,
                        start_period: str = None, end_period: str = None,
                        **kwargs) -> Union[PortfolioAsset, List[PortfolioAsset], None]:
        if start_period is None:
            start_period = self.__period_lowest
        if end_period is None:
            end_period = self.__period_highest()

        if 'name' in kwargs:
            start_period = pd.Period(start_period, freq='M')
            end_period = pd.Period(end_period, freq='M')

            name: str = kwargs['name']
            finsym_info = self.information(name=name)

            if finsym_info is None:
                return None
            if not isinstance(finsym_info, FinancialSymbol):
                raise ValueError('Unexpected type of financial symbol information')

            if currency is None:
                currency_enum: Currency = finsym_info.currency
            else:
                currency_enum = Currency.__dict__[currency.upper()]  # type: ignore

            allowed_security_types = {SecurityType.STOCK_ETF, SecurityType.MUT,
                                      SecurityType.CURRENCY, SecurityType.INDEX}
            assert finsym_info.security_type in allowed_security_types
            a = PortfolioAsset(symbol=finsym_info,
                               start_period=start_period, end_period=end_period, currency=currency_enum)
            return a
        elif 'names' in kwargs:
            names: List[str] = kwargs['names']
            assets: List[PortfolioAsset] = []
            for name in names:
                pa = self.portfolio_asset(name=name,
                                          start_period=start_period, end_period=end_period,
                                          currency=currency)
                if pa is None:
                    continue
                if not isinstance(pa, PortfolioAsset):
                    raise ValueError('Unexpected type of portfolio asset')
                assets.append(pa)
            return assets
        else:
            raise ValueError('Unexpected state of `kwargs`. Either `name`, or `names` should be given')
        pass

    @contract(
        assets='dict[N](str: float|int,>0), N>0',
    )
    def portfolio(self,
                  assets: Dict[str, float],
                  currency: str,
                  start_period: str = None, end_period: str = None) -> Portfolio:
        """
        :param assets: list of RostSber IDs. Supported security types: stock/ETF, MUT, Currency
        :param start_period: preferred period to start
        :param end_period: preferred period to end
        :param currency: common currency for all assets
        :return: returns instance of portfolio
        """
        if start_period is None:
            start_period = self.__period_lowest
        if end_period is None:
            end_period = self.__period_highest()

        names = list(assets.keys())
        assets_resolved = \
            self.portfolio_asset(names=names,
                                 start_period=str(pd.Period(start_period, freq='M') - 1),
                                 end_period=end_period,
                                 currency=currency)
        if not isinstance(assets_resolved, list):
            raise ValueError('`assets_resolved` should be list')
        asset2weight_dict: Dict[PortfolioAsset, float] = \
            {a: assets[a.symbol.identifier.format()] for a in assets_resolved}
        weights_sum: float = \
            np.abs(np.fromiter(asset2weight_dict.values(), dtype=float, count=len(asset2weight_dict)).sum())

        if np.abs(weights_sum - 1.) > 1e-3:
            asset2weight_dict = {a: (w / weights_sum) for a, w in asset2weight_dict.items()}

        start_period = pd.Period(start_period, freq='M')
        end_period = pd.Period(end_period, freq='M')
        currency_enum: Currency = Currency.__dict__[currency.upper()]  # type: ignore

        portfolio_instance = Portfolio(assets=list(asset2weight_dict.keys()),
                                       weights=list(asset2weight_dict.values()),
                                       start_period=start_period, end_period=end_period,
                                       currency=currency_enum)
        return portfolio_instance

    def available_names(self, **kwargs):
        """
        Returns the list of registered financial symbols names

        :param kwargs:
            either `namespace`, or `namespaces`, or nothing should be provided

            namespace (str): namespace of financial symbols

            namespaces (List[str]): a list of namespaces of financial symbols

            DEFAULT: returns the list of all registered namespaces

        :returns: (List[str]) list of financial symbols full names
        """
        if 'namespace' in kwargs:
            namespace = kwargs['namespace']
            return self.fin_syms_registry.get_all_infos(namespace)
        elif 'namespaces' in kwargs:
            namespaces = kwargs['namespaces']
            assert isinstance(namespaces, list)
            return [name
                    for namespace in namespaces
                    for name in self.available_names(namespace=namespace)]
        else:
            return self.fin_syms_registry.namespaces()

    def search(self, query: str, top=10):
        return self.__search.perform(query, top)

    def inflation(self, currency: str, kind: str,
                  end_period: str = None,
                  start_period: str = None, years_ago: int = None):
        currency_enum: Currency = Currency.__dict__[currency.upper()]  # type: ignore
        pc = self.portfolio_currency_factory.create(currency=currency_enum)
        if start_period:
            start_period = pd.Period(start_period, freq='M')
        elif years_ago is None:
            start_period = pc.period_min
        end_period = pd.Period(end_period, freq='M') if end_period else pc.period_max
        inflation_ts = pc.inflation(kind=kind,
                                    start_period=start_period, end_period=end_period,
                                    years_ago=years_ago)
        return inflation_ts



from enum import Enum, auto


class Currency(Enum):
    RUB = auto()
    USD = auto()
    EUR = auto()


class SecurityType(Enum):
    STOCK_ETF = auto()
    INDEX = auto()
    MUT = auto()
    CURRENCY = auto()
    INFLATION = auto()
    RATES = auto()


class Period(Enum):
    DAY = auto()
    MONTH = auto()
    DECADE = auto()



import re
from concurrent import futures
from concurrent.futures import ThreadPoolExecutor
from typing import List, Optional, Tuple, Iterator

from serum import inject

from ._sources.all_sources import AllSymbolSources
from ._sources.base_classes import FinancialSymbolsSource
from ._sources.registries import FinancialSymbolsRegistry
from .common.financial_symbol import FinancialSymbol
from .common.financial_symbol_id import FinancialSymbolId
from .common.financial_symbol_info import FinancialSymbolInfo


@inject
class _Search:
    sources: AllSymbolSources
    fin_syms_registry: FinancialSymbolsRegistry

    def __handle_quandl_info(self) -> List[str]:
        def func(x: FinancialSymbolInfo, src: FinancialSymbolsSource) -> str:
            fin_sym = src.fetch_financial_symbol(x.fin_sym_id.name)

            if fin_sym is None:
                return ''

            line = '{} {} {}'.format(fin_sym.name, fin_sym.exchange, fin_sym.short_name)
            line = re.sub(r'\s+', ' ', line.lower())

            self.id2sym.update({line: fin_sym})

            return line

        lines = [func(x, self.sources.quandl_source)
                 for x in self.sources.quandl_source.get_all_infos()]
        return lines

    def __handle_micex_stocks(self) -> List[str]:
        def func(x: FinancialSymbolInfo, src: FinancialSymbolsSource) -> str:
            fin_sym = src.fetch_financial_symbol(x.fin_sym_id.name)

            if fin_sym is None:
                return ''

            line = '{} {} {} {}'.format(fin_sym.name, fin_sym.exchange, fin_sym.isin, fin_sym.long_name)
            line = re.sub(r'\s+', ' ', line.lower())

            self.id2sym.update({line: fin_sym})

            return line

        lines = [func(x, self.sources.micex_stocks_source)
                 for x in self.sources.micex_stocks_source.get_all_infos()]
        return lines

    def __handle_mutru(self) -> List[str]:
        def func(x: FinancialSymbolInfo, src: FinancialSymbolsSource) -> str:
            fin_sym = src.fetch_financial_symbol(x.fin_sym_id.name)

            if fin_sym is None:
                return ''

            line = '{} {}'.format(fin_sym.name, fin_sym.short_name)
            line = re.sub(r'\s+', ' ', line.lower())

            self.id2sym.update({line: fin_sym})

            return line

        lines = [func(x, self.sources.mut_ru_source)
                 for x in self.sources.mut_ru_source.get_all_infos()]
        return lines

    def __init__(self):
        self.id2sym = {}

        pool = ThreadPoolExecutor(3)

        quandl_lines_fut = pool.submit(self.__handle_quandl_info)
        micex_stocks_lines_fut = pool.submit(self.__handle_micex_stocks)
        mutru_lines_fut = pool.submit(self.__handle_mutru)

        def handle_all_lines() -> List[str]:
            result = quandl_lines_fut.result() + micex_stocks_lines_fut.result() + mutru_lines_fut.result()
            return result

        self.lines_future: futures.Future[List[str]] = pool.submit(handle_all_lines)

    def _check_finsym_access(self, query: str) -> Optional[FinancialSymbol]:
        namespaces = self.fin_syms_registry.namespaces()

        starts_with_namespace = False
        for ns in namespaces:
            if query.startswith(ns + '/'):
                starts_with_namespace = True
                break
        if not starts_with_namespace:
            return None

        fsid = FinancialSymbolId.parse(query)
        return self.fin_syms_registry.get(fsid)

    @inject
    def perform(self, query: str, top: int) -> List[FinancialSymbol]:
        try:
            fs = self._check_finsym_access(query=query)
        except Exception:
            fs = None

        if fs is not None:
            return [fs]

        if not isinstance(query, str):
            raise ValueError('`query` should be string')
        if not isinstance(top, int):
            raise ValueError('`top` should be int')

        top = max(0, top)
        if not query or top == 0:
            return []

        query = re.sub(r'\s+', ' ', query.strip().lower())
        if len(query) == 0:
            return []

        lines: List[str] = self.lines_future.result()
        r: Iterator[Tuple[int, str]] = ((l.find(query), l) for l in lines)
        r = filter(lambda x: x[0] != -1, r)
        r_list = sorted(r, key=lambda x: '{:4d} {}'.format(x[0], x[1]))
        symbols: List[FinancialSymbol] = [self.id2sym[x[1]] for x in r_list[:top]]
        return symbols



import pandas as pd
import numpy as np
from enum import Enum
import copy
from collections.abc import Iterable

from .._settings import _MONTHS_PER_YEAR


class TimeSeriesKind(Enum):
    VALUES = 1
    DIFF = 2
    REDUCED_VALUE = 3
    YTD = 4
    CUMULATIVE = 5
    CURRENCY_RATE = 6
    VALUE = 7

    def __mul__(self, other):
        if isinstance(other, (int, float, complex)):
            return self
        if self == TimeSeriesKind.CURRENCY_RATE:
            return other
        if other == TimeSeriesKind.CURRENCY_RATE:
            return self
        if self != other:
            raise ValueError('kinds are incompatible')
        return self

    def __add__(self, other):
        if isinstance(other, (int, float, complex)):
            return self
        if self != other:
            raise ValueError('kinds are incompatible')
        return self

    def __radd__(self, other):
        if isinstance(other, (int, float, complex)):
            return self
        if self != other:
            raise ValueError('kinds are incompatible')
        return self

    def __rsub__(self, other):
        if isinstance(other, (int, float, complex)):
            return self
        if self != other:
            raise ValueError('kinds are incompatible')
        return self

    def __sub__(self, other):
        if isinstance(other, (int, float, complex)):
            return self
        if self != other:
            raise ValueError('kinds are incompatible')
        return self

    def __truediv__(self, other):
        if isinstance(other, (int, float, complex)):
            return self
        if self == TimeSeriesKind.CURRENCY_RATE:
            return other
        if other == TimeSeriesKind.CURRENCY_RATE:
            return self
        if self != other:
            raise ValueError('kinds are incompatible')
        return self

    def __pow__(self, power, modulo=None):
        return self

    def sqrt(self):
        return self

    def std(self):
        return self

    def mean(self):
        return self

    def sum(self):
        return self

    def prod(self):
        return self

    def cumprod(self):
        return TimeSeriesKind.CUMULATIVE

    def pct_change(self):
        return TimeSeriesKind.DIFF


class TimeSeries:
    def __init__(self, values, start_period: pd.Period, end_period: pd.Period, kind, freq='M'):
        if not isinstance(values, np.ndarray):
            raise ValueError('values should be numpy array')

        self._values = values
        self._size = self._values.size
        self._freq = freq
        self._kind = kind

        if self._freq == 'M':
            self._period_range = list(pd.period_range(start_period, end_period, freq='M'))
        elif self._freq == 'Y':
            rng = pd.period_range(start_period, end_period, freq='M')
            self._period_range = list(filter(lambda r: r.month == start_period.month, rng))
        else:
            raise ValueError('supported freq values: M, Y')
        self._start_period = min(self._period_range)
        self._end_period = max(self._period_range)

        if self.kind == TimeSeriesKind.DIFF or \
                self.kind == TimeSeriesKind.VALUES or \
                self.kind == TimeSeriesKind.CUMULATIVE:
            if self.size != (end_period - start_period).n + 1:
                raise ValueError('values and period range have different lengths')

        if self.kind == TimeSeriesKind.YTD:
            if self._start_period.month != 1:
                raise ValueError('start period month should be 1')
            if self._end_period.month != 1:
                raise ValueError('end period month should be 1')
            if len(values) != self._end_period.year - self._start_period.year + 1:
                raise ValueError('values len should be equal to full years count')

        if self.kind == TimeSeriesKind.REDUCED_VALUE:
            if self.size > 1:
                raise ValueError('size is greater than 1')

    def __validate(self, time_series):
        if self._start_period != time_series.start_period:
            raise ValueError('start periods are incompatible')
        if self._end_period != time_series.end_period:
            raise ValueError('end periods are incompatible')

    @property
    def values(self):
        return self._values

    @property
    def value(self):
        if self.kind in {TimeSeriesKind.REDUCED_VALUE, TimeSeriesKind.VALUE}:
            return self._values[0]
        raise ValueError('incorrect `kind` to get value')

    @property
    def kind(self):
        return self._kind

    @property
    def start_period(self):
        return self._start_period

    @property
    def end_period(self):
        return self._end_period

    @property
    def period_size(self):
        return (self._end_period - self._start_period).n + 1

    def period_range(self):
        return copy.deepcopy(self._period_range)

    @property
    def size(self):
        return self._size

    def pct_change(self):
        if len(self._values) < 2:
            raise ValueError('`value` length should be >= 2')
        vals = np.diff(self._values) / self._values[:-1]
        return TimeSeries(values=vals,
                          start_period=self._start_period + 1, end_period=self._end_period,
                          freq=self._freq,
                          kind=self.kind.pct_change())

    def apply(self, fun, *args):
        if len(args) == 0:
            ts = TimeSeries(values=fun(self._values),
                            start_period=self._start_period, end_period=self._end_period,
                            freq=self._freq,
                            kind=fun(self.kind))
            return ts
        else:
            other = args[0]
            if isinstance(other, TimeSeries):
                self.__validate(other)
                ts = TimeSeries(values=fun(self._values, other._values),
                                start_period=self._start_period, end_period=self._end_period,
                                freq=self._freq,
                                kind=fun(self.kind, other.kind))
                return ts
            elif isinstance(other, (int, float, complex)):
                ts = TimeSeries(values=fun(self._values, other),
                                start_period=self._start_period, end_period=self._end_period,
                                freq=self._freq,
                                kind=self._kind)
                return ts
            else:
                raise ValueError('argument has incompatible type')

    def reduce(self, fun):
        return TimeSeries(values=np.array([fun(self._values)]),
                          start_period=self._start_period, end_period=self._end_period,
                          freq=self._freq,
                          kind=TimeSeriesKind.REDUCED_VALUE)

    def __mul__(self, other):
        return self.apply(lambda x, y: x * y, other)

    def __add__(self, other):
        if isinstance(other, Iterable):
            vals_new = np.append(self._values, other)

            if self._freq == 'Y':
                end_period_new = self.end_period + len(list(other)) * _MONTHS_PER_YEAR
            elif self._freq == 'M':
                end_period_new = self.end_period + len(list(other))
            else:
                raise ValueError('supported freq values: M, Y')

            ts = TimeSeries(values=vals_new,
                            start_period=self.start_period, end_period=end_period_new,
                            freq=self._freq,
                            kind=self._kind)
            return ts

        return self.apply(lambda x, y: x + y, other)

    def __radd__(self, other):
        if isinstance(other, Iterable):
            vals_new = np.append(other, self._values)

            if self._freq == 'Y':
                start_period_new = self.start_period - len(list(other)) * _MONTHS_PER_YEAR
            elif self._freq == 'M':
                start_period_new = self.start_period - len(list(other))
            else:
                raise ValueError('supported freq values: M, Y')

            ts = TimeSeries(values=vals_new,
                            start_period=start_period_new, end_period=self.end_period,
                            freq=self._freq,
                            kind=self._kind)
            return ts

        return self.apply(lambda x, y: y + x, other)

    def __rsub__(self, other):
        return self.apply(lambda x, y: y - x, other)

    def __rmul__(self, other):
        return self.apply(lambda x, y: y * x, other)

    def __sub__(self, other):
        return self.apply(lambda x, y: x - y, other)

    def __truediv__(self, other):
        return self.apply(lambda x, y: x / y, other)

    def __getitem__(self, key):
        if isinstance(key, slice):
            if not (key.step is None or key.step == 1):
                raise ValueError('step value is not supported: {}'.format(key.step))
            pr = pd.period_range(start=self._start_period, end=self._end_period, freq='M')
            pr = pr[key.start:key.stop]
            ts = TimeSeries(values=self._values[key.start:key.stop],
                            start_period=pr.min(), end_period=pr.max(),
                            freq=self._freq,
                            kind=self._kind)
            return ts
        elif isinstance(key, int):
            if key > 0:
                p = self._start_period + key
            else:
                p = self._end_period + key

            ts = TimeSeries(values=np.array([self._values[key]]),
                            start_period=p, end_period=p,
                            freq=self._freq,
                            kind=TimeSeriesKind.VALUE)
            return ts
        else:
            raise ValueError('Type of `key` is not supported')

    def __pow__(self, power, modulo=None):
        return self.apply(lambda x: x ** power)

    def sqrt(self):
        return self.apply(lambda x: np.sqrt(x))

    def std(self):
        return self.reduce(lambda x: x.std())

    def mean(self):
        return self.reduce(lambda x: x.mean())

    def sum(self):
        return self.reduce(lambda x: x.sum())

    def prod(self):
        return self.reduce(lambda x: x.prod())

    def cumprod(self):
        return self.apply(lambda x: x.cumprod())

    def ytd(self):
        if self._kind != TimeSeriesKind.DIFF:
            raise ValueError('incorrect kind')
        if self._freq != 'M':
            raise ValueError('incorrect frequency')

        drop_first_count = None if self.start_period.month == 1 else _MONTHS_PER_YEAR - self.start_period.month + 1
        drop_last_count = None if self.end_period.month == _MONTHS_PER_YEAR else -self.end_period.month
        if drop_first_count and drop_last_count and drop_first_count + (-drop_last_count) == self.size:
            msg = "YTD for the current dates range ({} - {}) is empty".format(self.start_period, self.end_period)
            raise ValueError(msg)
        values_full_yearly = self[drop_first_count:drop_last_count]
        values_full_yearly_splits = \
            np.split(values_full_yearly.values,
                     values_full_yearly.end_period.year - values_full_yearly.start_period.year + 1)
        values_ytd_ts = [((splt + 1.).prod() - 1.) for splt in values_full_yearly_splits]
        values_ytd = TimeSeries(values=np.array(values_ytd_ts),
                                start_period=values_full_yearly.start_period,
                                end_period=values_full_yearly.end_period,
                                freq='Y',
                                kind=TimeSeriesKind.YTD)
        return values_ytd

    def __repr__(self):
        return 'TimeSeries(start_period={}, end_period={}, kind={}, values={}'.format(
            self._start_period, self._end_period, self._kind, self._values
        )



from serum import Context

from ._instance import Yapo
from ._sources.all_sources import AllSymbolSources

with Context(AllSymbolSources):
    yapo_instance = Yapo()
information = yapo_instance.information
portfolio = yapo_instance.portfolio
portfolio_asset = yapo_instance.portfolio_asset
available_names = yapo_instance.available_names
search = yapo_instance.search
inflation = yapo_instance.inflation



import datetime as dtm
from collections import namedtuple
from typing import Optional, Callable

import pandas as pd
from dateutil import relativedelta

from ..common.enums import Period, Currency, SecurityType
from ..common.financial_symbol_id import FinancialSymbolId


class ValuesFetcher:

    _PeriodRange = namedtuple('PeriodRange', 'start, end')

    def __init__(self, values_func, period_min, period_max):
        self._values_func = values_func
        self._period_min = period_min
        self._period_max = period_max
        self._current_period_start = None
        self._current_period_end = None
        self._values = None

    @property
    def _period_range(self):
        return self._PeriodRange(self._current_period_start, self._current_period_end)

    def _fetch(self, start_period: pd.Period, end_period: pd.Period):
        start_period = max(start_period, self._period_min)
        end_period = min(end_period, self._period_max)

        if self._current_period_start is None or \
                (start_period < self._current_period_start and end_period > self._current_period_end):
            self._current_period_start = start_period
            self._current_period_end = end_period
            self._values = self._values_func(start_period, end_period)

        elif start_period >= self._current_period_start and end_period <= self._current_period_end:
            pass

        elif start_period < self._current_period_start:
            self._values = self._values_func(start_period, self._current_period_start - 1).append(self._values)
            self._current_period_start = start_period

        elif end_period > self._current_period_end:
            self._values = self._values.append(self._values_func(self._current_period_end + 1, end_period))
            self._current_period_end = end_period

        else:
            pass

        periods = self._values.date.dt.to_period(freq='M')
        df = self._values[(start_period <= periods) & (periods <= end_period)]
        return df.copy()


class FinancialSymbol:
    def __init__(self,
                 identifier: FinancialSymbolId,
                 values: Callable[[pd.Period, pd.Period], pd.DataFrame],
                 start_period: pd.Period,
                 end_period: pd.Period,
                 adjusted_close: bool,
                 currency: Currency,
                 period: Period,
                 security_type: SecurityType,
                 isin: Optional[str] = None,
                 short_name: Optional[str] = None,
                 long_name: Optional[str] = None,
                 exchange: Optional[str] = None):
        self.identifier = identifier
        self.__values_fetcher = ValuesFetcher(
            values_func=values,
            period_min=pd.Period(start_period, freq='M'),
            period_max=pd.Period(end_period, freq='M')
        )
        self.isin = isin
        self.short_name = short_name
        self.long_name = long_name
        self.exchange = exchange
        self.currency = currency
        self.security_type = security_type
        self.start_period = start_period
        self.end_period = end_period
        self.period = period
        self.adjusted_close = adjusted_close

    def values(self, start_period: pd.Period, end_period: pd.Period) -> pd.DataFrame:
        vals = self.values_fetcher._fetch(start_period=start_period, end_period=end_period)

        if self.period == Period.DAY:
            # we are interested in day-time data as follows
            # - if there is no data for month or more (ticker is dead, probably), we drop all data for
            #   the last available period
            # - we drop data for the period of the current month
            # - for every period we take the value that is last in each month

            if 'period' not in vals.columns:
                vals['period'] = vals['date'].dt.to_period('M')

            month_ago = pd.Period(dtm.datetime.now() - relativedelta.relativedelta(months=1), freq='D')
            if self.end_period < month_ago:
                vals = vals[vals['period'] < pd.Period(self.end_period, freq='M')]
            indicator__not_current_period = vals['period'] != pd.Period.now(freq='M')
            indicator__lastdate_indices = vals['period'] != vals['period'].shift(1)
            vals = vals[indicator__lastdate_indices & indicator__not_current_period].copy()
            del vals['date']
        elif self.period == Period.MONTH:
            del vals['date']
        elif self.period == Period.DECADE:
            vals = vals[vals['date'].dt.day == 3]
            vals['date'] = vals['date'].apply(lambda p: pd.Period(p, freq='M'))
            vals.rename(columns={'date': 'period'}, inplace=True)
        else:
            raise Exception('Unexpected type of `period`')

        vals.sort_values(by='period', ascending=True, inplace=True)
        return vals

    @property
    def values_fetcher(self) -> ValuesFetcher:
        return self.__values_fetcher

    @property
    def name(self) -> str:
        return self.identifier.name

    @property
    def namespace(self) -> str:
        return self.identifier.namespace

    @property
    def identifier_str(self) -> str:
        return self.identifier.format()

    def __repr__(self) -> str:
        from pprint import pformat
        return '{}\n{}'.format(self.identifier_str, pformat(vars(self)))



from pprint import pformat
from typing import Optional

import pandas as pd

from ..common.financial_symbol import FinancialSymbol
from ..common.financial_symbol_id import FinancialSymbolId
from ..common.financial_symbol_info import FinancialSymbolInfo


class FinancialSymbolsSource:
    def __init__(self, namespace):
        self.namespace = namespace

    def fetch_financial_symbol(self, name: str) -> Optional[FinancialSymbol]:
        raise Exception('should not be called')

    def __repr__(self):
        return pformat(vars(self))

    def get_all_infos(self):
        raise Exception('should not be called')


class SingleFinancialSymbolSource(FinancialSymbolsSource):
    def __extract_values(self, start_period: pd.Period, end_period: pd.Period) -> pd.DataFrame:
        df = self.__values_fetcher()
        df['date'] = pd.to_datetime(df['date'])
        return df[(df['date'].dt.to_period('M') >= start_period) &
                  (df['date'].dt.to_period('M') <= end_period)].copy()

    def __init__(self, values_fetcher, namespace, name, start_period, end_period,
                 isin=None,
                 short_name=None,
                 long_name=None,
                 exchange=None,
                 currency=None,
                 security_type=None,
                 period=None,
                 adjusted_close=None):
        super().__init__(namespace)
        self.name = name
        self.__values_fetcher = values_fetcher
        self.financial_symbol = \
            FinancialSymbol(identifier=FinancialSymbolId(namespace=self.namespace, name=self.name),
                            values=self.__extract_values,
                            isin=isin,
                            short_name=short_name,
                            long_name=long_name,
                            start_period=start_period,
                            end_period=end_period,
                            exchange=exchange,
                            currency=currency,
                            security_type=security_type,
                            period=period,
                            adjusted_close=adjusted_close)

    def fetch_financial_symbol(self, name: str) -> Optional[FinancialSymbol]:
        return self.financial_symbol if name == self.name else None

    def get_all_infos(self):
        fin_sym_info = FinancialSymbolInfo(
            fin_sym_id=self.financial_symbol.identifier,
            short_name=self.financial_symbol.short_name
        )
        return [fin_sym_info]



import datetime as dtm
from functools import lru_cache
from typing import Optional, Callable

import pandas as pd
from serum import singleton

from .base_classes import SingleFinancialSymbolSource, FinancialSymbolsSource
from .._settings import data_url, change_column_name
from ..common.enums import Currency, SecurityType, Period
from ..common.financial_symbol import FinancialSymbol
from ..common.financial_symbol_id import FinancialSymbolId
from ..common.financial_symbol_info import FinancialSymbolInfo


@singleton
class CbrTopRatesSource(SingleFinancialSymbolSource):
    def _load_rates(self):
        df = pd.read_csv('{}cbr_deposit_rate/data.csv'.format(data_url), sep='\t')
        df.sort_values(by='decade', inplace=True)
        df.rename(columns={'close': change_column_name, 'decade': 'date'},
                  inplace=True)
        return df

    def _load_dates(self, kind):
        index = pd.read_csv('{}cbr_deposit_rate/__index.csv'.format(data_url), sep='\t')
        period_str = index[kind][0]
        return pd.Period(period_str, freq='M')

    def __init__(self):
        super().__init__(
            namespace='cbr',
            name='TOP_rates',
            values_fetcher=lambda: self._load_rates(),
            start_period=self._load_dates(kind='date_start'),
            end_period=self._load_dates(kind='date_end'),
            long_name='Динамика максимальной процентной ставки (по вкладам в российских рублях)',
            currency=Currency.RUB,
            security_type=SecurityType.RATES,
            period=Period.DECADE,
            adjusted_close=False,
        )


@singleton
class CbrCurrenciesSource(FinancialSymbolsSource):
    def __init__(self):
        super().__init__(namespace='cbr')
        self.url_base = data_url + 'currency/'
        self.index = pd.read_csv(self.url_base + '__index.csv', sep='\t', index_col='name')
        self.__short_names = {
            Currency.RUB: 'Рубль РФ',
            Currency.USD: 'Доллар США',
            Currency.EUR: 'Евро',
        }
        self._currency_min_date = {
            Currency.RUB.name: pd.Period('1990', freq='D'),
            Currency.USD.name: pd.Period('1913', freq='D'),
            Currency.EUR.name: pd.Period('1996', freq='D'),
        }

    @lru_cache(maxsize=512)
    def __currency_values(self, name: str) -> Callable[[pd.Period, pd.Period], pd.DataFrame]:
        def func(start_period: pd.Period, end_period: pd.Period) -> pd.DataFrame:
            start_period = max(start_period,
                               pd.Period(self._currency_min_date[name], freq='M'))
            end_period = min(end_period, pd.Period.now(freq='M'))
            date_range = pd.date_range(start=start_period.to_timestamp(),
                                       end=(end_period + 1).to_timestamp(),
                                       freq='D')
            df = pd.DataFrame({'date': date_range, 'close': 1.0})

            df['period'] = df['date'].dt.to_period('M')
            df_new = df[(start_period <= df['period']) & (df['period'] <= end_period)].copy()
            return df_new

        return func

    def fetch_financial_symbol(self, name: str) -> Optional[FinancialSymbol]:
        name = name.upper()
        currency = Currency.__dict__.get(name)  # type: ignore

        if currency is None:
            return None

        fs = FinancialSymbol(
            identifier=FinancialSymbolId(namespace='cbr', name=name),
            values=self.__currency_values(name),
            short_name=self.__short_names[currency],
            start_period=self._currency_min_date[currency.name],
            end_period=pd.Period(dtm.datetime.now(), freq='D'),
            currency=currency,
            security_type=SecurityType.CURRENCY,
            period=Period.DAY,
            adjusted_close=True,
        )
        return fs

    def get_all_infos(self):
        return [
            FinancialSymbolInfo(
                fin_sym_id=FinancialSymbolId(self.namespace, short_name.name),
                short_name=short_name,
            ) for short_name in self.__short_names.keys()
        ]



import pandas as pd
from contracts import contract
from serum import inject, singleton

from .._settings import _MONTHS_PER_YEAR
from .._sources.inflation_source import InflationSource
from .._sources.single_financial_symbol_source import CbrCurrenciesSource
from ..common.enums import Currency
from ..common.financial_symbol import FinancialSymbol
from ..common.time_series import TimeSeries, TimeSeriesKind


@inject
class PortfolioCurrency:
    __inflation_source: InflationSource
    __cbr_currencies_source: CbrCurrenciesSource

    def __init__(self, currency: Currency):
        self._currency = currency

        inflation_symbol = self.__inflation_source.fetch_financial_symbol(currency.name)
        if inflation_symbol is None:
            raise ValueError('Inflation symbol for the `name`={} is not found'.format(currency.name))
        self._inflation_symbol: FinancialSymbol = inflation_symbol

        currency_symbol = self.__cbr_currencies_source.fetch_financial_symbol(currency.name)
        if currency_symbol is None:
            raise ValueError('Currency symbol for the `name`={} is not found'.format(currency.name))
        self._currency_symbol: FinancialSymbol = currency_symbol
        self._period_min: pd.Period = max(
            self.inflation_start_period - 1,
            self._currency_symbol.start_period.asfreq(freq='M'),
        )
        self._period_max: pd.Period = min(
            self.inflation_end_period,
            self._currency_symbol.end_period.asfreq(freq='M'),
        )

    @property
    def period_min(self) -> pd.Period:
        return self._period_min

    @property
    def period_max(self) -> pd.Period:
        return self._period_max

    @property
    def inflation_start_period(self) -> pd.Period:
        return self._inflation_symbol.start_period.asfreq(freq='M')

    @property
    def inflation_end_period(self) -> pd.Period:
        return self._inflation_symbol.end_period.asfreq(freq='M')

    @property
    def value(self) -> Currency:
        return self._currency

    @contract(
        kind='str',
        years_ago='int,>0|None',
    )
    def inflation(self, kind,
                  end_period: pd.Period,
                  start_period: pd.Period = None,
                  years_ago: int = None) -> TimeSeries:
        """
        Computes the properly reduced inflation for the currency

        :param start_period:
            the period from which the inflation is calculated
        :param end_period:
            the period to which the inflation is calculated
            is computed automatically if `years_ago` is provided
        :param years_ago:
            years back from `period_max` to calculate inflation

        :param kind:
            cumulative - cumulative inflation

            a_mean - arithmetic mean of inflation

            g_mean - geometric mean of inflation

            yoy - Year on Year inflation

            values - raw values of inflation
        :return:
        """
        if (years_ago is None) == (start_period is None):
            raise ValueError('either `start_period` or `years_ago` should be provided')

        if years_ago is not None:
            start_period = end_period - years_ago * _MONTHS_PER_YEAR + 1

        inflation_values = self._inflation_symbol.values(start_period=start_period,
                                                         end_period=end_period)

        inflation_ts = TimeSeries(values=inflation_values.value.values,
                                  start_period=inflation_values.period.min(),
                                  end_period=inflation_values.period.max(),
                                  kind=TimeSeriesKind.DIFF)

        def __cumulative():
            return (inflation_ts + 1.).prod() - 1.

        if kind == 'cumulative':
            return __cumulative()
        elif kind == 'yoy':
            return inflation_ts.ytd()
        elif kind == 'cumulative_series':
            return (inflation_ts + 1.).cumprod() - 1.
        elif kind == 'a_mean':
            inflation_amean = inflation_ts.mean()
            return inflation_amean
        elif kind == 'g_mean':
            years_total = ((end_period - start_period).n + 1) / _MONTHS_PER_YEAR
            inflation_gmean = (__cumulative() + 1.) ** (1 / years_total) - 1.
            return inflation_gmean
        elif kind == 'values':
            return inflation_ts
        else:
            raise ValueError('inflation kind is not supported: {}'.format(kind))

    def __repr__(self) -> str:
        currency_repr = 'Currency({})'.format(self._currency.name)
        return currency_repr


@singleton
class PortfolioCurrencyFactory:

    @staticmethod
    def create(currency: Currency) -> PortfolioCurrency:
        return PortfolioCurrency(currency=currency)



from typing import Optional

from ..common.time_series import TimeSeries, TimeSeriesKind
from .._settings import _MONTHS_PER_YEAR

import numpy as np
import pandas as pd


def compute(cbr_top10_rates: pd.DataFrame) -> TimeSeries:
    index_total: Optional[TimeSeries] = None
    start_period: pd.Period = cbr_top10_rates['period'].min()
    end_period: pd.Period = cbr_top10_rates['period'].max()
    for month_idx in range(_MONTHS_PER_YEAR):
        rates_yearly = cbr_top10_rates['rate'].values[month_idx::_MONTHS_PER_YEAR]
        index_part = TimeSeries(values=np.repeat(rates_yearly, _MONTHS_PER_YEAR)[:len(cbr_top10_rates) - month_idx],
                                start_period=start_period + month_idx + 1,
                                end_period=end_period + 1,
                                kind=TimeSeriesKind.VALUES)
        index_part = (1 + index_part / _MONTHS_PER_YEAR).cumprod()
        index_total = index_part if index_total is None else index_total[1:] + index_part

    if index_total is None:
        raise Exception('`index_total` should not be `None`')
    result: TimeSeries = index_total
    result = (result.pct_change() + 1.).cumprod() - 1.
    result = [0] + result
    result = (result + 1) * 100
    return result



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['yapo', 'yapo._index', 'yapo._portfolio', 'yapo._sources', 'yapo.common']

package_data = \
{'': ['*']}

install_requires = \
['PyContracts>=1.8,<2.0',
 'numpy>=1.16,<2.0',
 'pandas>=0.24,<0.25',
 'quandl>=3.4,<4.0',
 'serum>=5.1,<6.0']

setup_kwargs = {
    'name': 'yapo',
    'version': '0.2.0',
    'description': 'Flexible and easy-to-use Python library for analysis & manipulation with financial & economic data',
    'long_description': '|Travis|_ |Gitter|_ |Pepy|_ |Codecov|_\n\n.. |Travis| image:: https://travis-ci.org/okama-io/yapo.svg?branch=master\n.. _Travis: https://travis-ci.org/okama-io/yapo\n\n.. |Gitter| image:: https://badges.gitter.im/okama-io/community.svg\n.. _Gitter: https://gitter.im/okama-io/community\n\n.. |Pepy| image:: https://pepy.tech/badge/yapo\n.. _Pepy: https://pepy.tech/badge/yapo\n\n.. |Codecov| image:: https://codecov.io/gh/okama-io/yapo/branch/master/graph/badge.svg\n.. _Codecov: https://codecov.io/gh/okama-io/yapo\n\n**YAPO** |--| a flexible and easy-to-use Python 3.6+ library for analysis &\nmanipulation with financial & economic data\n\n.. contents:: Contents of this Document\n\nIntroduction\n============\n\n``yapo`` is a Python library developed to solve quantitative finance and\ninvestments tasks. Additionally, it has the broader goal to become the most\nuseful and flexible open sourced tool for financial data analysis available in\npopular programming languages.\n\nApplications\n============\n\nUseful applications of ``yapo`` in the community developed are as follows:\n\n* Fully-functional ``Angular2+``-based web-application `okama.io`_\n* `tell us more <okama-discourse_>`_ if you know any\n\nThe Ecosystem\n=============\n\nThe ecosystem around the library consists of:\n\n* The Python package\n* Financial and Economic data for several markets\n* ``Angular2+`` web-application `okama.io`_\n* `Financial terms glossary <okama-glossary_>`_\n* `The community <okama-discourse_>`_\n\nMain Features\n=============\n\n* ``[x]`` ``TimeSeries`` to verify correctness of financial data manipulations\n* ``[x]`` Error-free manipulations with financial data checked by tests and active community\n* ``[x]`` Asset analysis tools for asset correlations and main performance indicators\n* ``[ ]`` Portfolio analysis tools for asset class allocation and portfolio backtesting\n* ``[ ]`` Portfolio optimization and efficient frontier visualization\n* ``[ ]`` Monte Carlo Simulation for financial assets and investment portfolios\n* ``[ ]`` Bonds key properties calculations\n* ``[x]`` Access to financial data from different stock markets: EOD close, adjusted close, currency rates, inflation\n* ``[x]`` Financial and Economic data with API with GraphQL data access\n\nFinancial and Economic data freely available\n============================================\n* ``[x]`` EOD adjusted close for NYSE and NASDAQ stocks and ETF\n* ``[x]`` EOD close for Moscow Exchange stocks and ETF\n* ``[x]`` EOD close for Russian open-end funds\n* ``[ ]`` EOD close for BSE and NSE stocks and ETF (India)\n* ``[x]`` EOD for main stock and bond Indexes\n* ``[ ]`` Bonds data for Moscow Exchange-traded securities: EOD close, coupons, maturity\n* ``[x]`` Exchange Rates for USD, EUR, RUB\n* ``[ ]`` Exchange Rates for Bitcoin [BTC], Ethereum [ETH], Binance Coin [BNB] and other cryptocurrencies\n* ``[x]`` Inflation for US, EU, and Russia\n* ``[x]`` Key interest rates for US, EU, and Russia\n* ``[x]`` History of deposit rates for top 10 banks of Russia\n\nInstallation\n============\n\nTODO\n\nJupyter Notebooks\n=================\n\nTODO\n\nDependencies\n============\n\nThe library dependencies are listed at\n`pyproject.toml <https://github.com/okama-io/yapo/blob/master/pyproject.toml>`_ under\n``[tool.poetry.dependencies]`` section.\n\nDiscussion, Development, and Getting Help\n=========================================\n\n- The development discussion takes place at `the GitHub repo\n  <yapo-github-issues_>`_. We encourage you to report issues using `the Github\n  tracker <yapo-github-issues_>`_. We welcome all kinds of issues related to\n  correctness, documentation, performance, and feature requests.\n- `The community forum <okama-discourse_>`_ can also be used for general\n  questions and discussions.\n- Finally, the `Gitter channel <Gitter_>`_ is available for the development\n  related questions.\n\nContributing\n============\n\nAll contributions, bug reports, bug fixes, documentation improvements,\nenhancements, frontend implementation, and ideas are welcomed and the subject\nto discuss. Simple ways to start contributing immediately:\n\n- Browse the issue tracker to find issues that interest you\n- Read the source code and improve the documentation or address TODOs\n- Improve the example library and tutorials\n- Bug reports are an important part of making the library more stable\n- Run the library through `the okama.io frontend <okama.io_>`_ and suggest\n  improvements in design, UI, and functionality\n\nThe code is hosted at `GitHub <yapo-github_>`_. You need an GitHub account\nwhich is free to contribute to the project. We use git for the version control\nto enable distributed work on the project.\n\nContributions should be submitted as a pull request. A member of the\ndevelopment team will review the pull request and guide you through the\ncontributing process.\n\nFeel free to ask questions at `the community <okama-discourse_>`_.\n\nLicense\n=======\n\n`MIT <license_>`_\n\n.. |--| unicode:: U+2013\n.. _okama.io: https://okama.io/\n.. _okama-glossary: https://okama.io/#/glossary\n.. _okama-discourse: http://community.okama.io\n.. _yapo-github: https://github.com/okama-io/yapo\n.. _yapo-github-issues: https://github.com/okama-io/yapo/issues\n.. _license: https://github.com/okama-io/yapo/blob/master/LICENSE\n',
    'author': 'Alexander Myltsev',
    'author_email': None,
    'url': 'https://github.com/okama-io/yapo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from serum import singleton, inject
import pandas as pd

from .base_classes import FinancialSymbolsSource
from .single_financial_symbol_source import CbrTopRatesSource
from ..common.financial_symbol_info import FinancialSymbolInfo
from ..common.enums import Currency, SecurityType, Period
from ..common.financial_symbol_id import FinancialSymbolId
from ..common.financial_symbol import FinancialSymbol
from .._index.okid10 import compute as compute_top10


@singleton
@inject
class OkamaSource(FinancialSymbolsSource):
    cbr_top_rates_source: CbrTopRatesSource

    def __init__(self):
        super().__init__(namespace='index')

        self.cbr_top10_sym = self.cbr_top_rates_source.fetch_financial_symbol('TOP_rates')
        self.okid10_name = 'OKID10'

    def fetch_financial_symbol(self, name: str) -> Optional[FinancialSymbol]:
        if name != self.okid10_name:
            return None

        def fetch_values(start_period: pd.Period, end_period: pd.Period) -> pd.DataFrame:
            vals = self.cbr_top10_sym.values(
                start_period=self.cbr_top10_sym.start_period,
                end_period=end_period,
            )
            index_vals = compute_top10(vals)
            drop_first = max(0, (start_period - index_vals.start_period).n)
            index_vals = index_vals[drop_first:]
            df = pd.DataFrame({'date': pd.date_range(index_vals.start_period.to_timestamp(),
                                                     (index_vals.end_period + 1).to_timestamp(),
                                                     freq='M'),
                               'period': pd.period_range(index_vals.start_period,
                                                         index_vals.end_period,
                                                         freq='M'),
                               'close': index_vals.values})
            return df
        sym = FinancialSymbol(identifier=FinancialSymbolId(namespace=self.namespace, name=name),
                              values=fetch_values,
                              short_name='Okama TOP10 Index',
                              long_name='Okama TOP10 Index',
                              start_period=self.cbr_top10_sym.start_period,
                              end_period=self.cbr_top10_sym.end_period,
                              currency=Currency.RUB,
                              security_type=SecurityType.INDEX,
                              period=Period.MONTH,
                              adjusted_close=True)
        return sym

    def get_all_infos(self):
        fsi = FinancialSymbolInfo(
            fin_sym_id=FinancialSymbolId(self.namespace, self.okid10_name),
            short_name=self.okid10_name,
        )
        infos = [fsi]
        return infos



import os
from functools import lru_cache
from typing import Optional, Callable

import pandas as pd
import quandl
from serum import singleton

from .base_classes import FinancialSymbolsSource
from .._settings import data_url
from ..common.enums import Currency, SecurityType, Period
from ..common.financial_symbol import FinancialSymbol
from ..common.financial_symbol_id import FinancialSymbolId
from ..common.financial_symbol_info import FinancialSymbolInfo


@singleton
class QuandlSource(FinancialSymbolsSource):
    def __init__(self):
        super().__init__(namespace='ny')

        quandl_key_name = 'QUANDL_KEY'
        if quandl_key_name in os.environ:
            quandl.ApiConfig.api_key = os.environ[quandl_key_name]
            self._available = True
            self.url_base = data_url + 'quandl/'
            self.index = pd.read_csv(self.url_base + '__index.csv', sep='\t', index_col='name',
                                     parse_dates=['date_start', 'date_end'])
            self.index['date_start'] = self.index['date_start'].dt.to_period(freq='D')
            self.index['date_end'] = self.index['date_end'].dt.to_period(freq='D')
        else:
            self._available = False
            self.index = pd.DataFrame({
                'name': [],
                'date_start': [],
                'date_end': [],
                'short_name': [],
                'exchange': [],
            })

    @lru_cache(maxsize=512)
    def __extract_values(self, name: str) -> Callable[[pd.Period, pd.Period], pd.DataFrame]:
        name = name.replace('.', '_')

        def func(start_period: pd.Period, end_period: pd.Period) -> pd.DataFrame:
            nonlocal name
            df = quandl.get('EOD/{}.11'.format(name),
                            start_date=start_period.to_timestamp().strftime('%Y-%m-%d'),
                            end_date=end_period.to_timestamp('M').strftime('%Y-%m-%d'),
                            collapse='monthly')
            df_res = pd.DataFrame()
            df_res['close'] = df['Adj_Close']
            df_res['date'] = df_res.index
            return df_res

        return func

    def fetch_financial_symbol(self, name: str) -> Optional[FinancialSymbol]:
        if not self._available:
            return None

        if name in self.index.index:
            row = self.index.loc[name]
            symbol = FinancialSymbol(identifier=FinancialSymbolId(namespace=self.namespace, name=name),
                                     values=self.__extract_values(name),
                                     exchange=row['exchange'],
                                     short_name=row['short_name'],
                                     start_period=row['date_start'],
                                     end_period=row['date_end'],
                                     currency=Currency.USD,
                                     security_type=SecurityType.STOCK_ETF,
                                     period=Period.DAY,
                                     adjusted_close=True)
            return symbol
        return None

    def get_all_infos(self):
        if not self._available:
            return []

        infos = []
        for idx, row in self.index.iterrows():
            fsi = FinancialSymbolId(self.namespace, str(idx))
            fs_info = FinancialSymbolInfo(fin_sym_id=fsi, short_name=idx)
            infos.append(fs_info)
        return infos



import copy
import datetime as dtm
from textwrap import dedent
from typing import List, Dict

import dateutil.relativedelta
import numpy as np
import pandas as pd
from contracts import contract
from serum import inject

from .currency import PortfolioCurrencyFactory
from .._settings import _MONTHS_PER_YEAR
from .._sources.registries import CurrencySymbolsRegistry
from ..common.enums import Currency, Period
from ..common.financial_symbol import FinancialSymbol
from ..common.time_series import TimeSeries, TimeSeriesKind


@inject
class PortfolioAsset:
    currency_symbols_registry: CurrencySymbolsRegistry
    portfolio_currency_factory: PortfolioCurrencyFactory

    def __init__(self, symbol: FinancialSymbol,
                 start_period: pd.Period, end_period: pd.Period, currency: Currency,
                 portfolio: 'Portfolio' = None,
                 weight: int = None):
        if (end_period - start_period).n < 2:
            raise ValueError('period range should be at least 2 months')

        self.symbol = symbol
        self.currency = self.portfolio_currency_factory.create(currency=currency)
        self._portfolio = portfolio
        self._weight = weight

        datetime_now = dtm.datetime.now()
        if (datetime_now + dtm.timedelta(days=1)).month == datetime_now.month:
            datetime_now -= dateutil.relativedelta.relativedelta(months=1)
        period_now = pd.Period(datetime_now, freq='M')
        self._period_min = max(
            pd.Period(self.symbol.start_period, freq='M'),
            self.currency.period_min,
            start_period,
        )
        self._period_max = min(
            pd.Period(self.symbol.end_period, freq='M'),
            period_now,
            self.currency.period_max,
            end_period,
        )
        if self._period_min >= self._period_max:
            raise ValueError('`self._period_min` must not be >= `self._period_max`')

        currency_conversion_rate = self.__currency_conversion_rate(currency_to=self.currency.value)
        self._period_min = max(self._period_min, currency_conversion_rate.start_period)
        self._period_max = min(self._period_max, currency_conversion_rate.end_period)

        self.__values = self.__transform_values_according_to_period()

    @property
    def portfolio(self):
        return self._portfolio

    @property
    def weight(self):
        return self._weight

    def __transform_values_according_to_period(self):
        vals = self.symbol.values(start_period=self._period_min, end_period=self._period_max)
        if len(vals['period']) > 0:
            self._period_min = max(self._period_min, vals['period'].min())
            self._period_max = min(self._period_max, vals['period'].max())
        # TODO: okama_dev-98
        if self.symbol.period == Period.DECADE:
            ts = TimeSeries(values=vals['rate'].values,
                            start_period=self._period_min, end_period=self._period_max,
                            kind=TimeSeriesKind.DIFF)
        else:
            ts = TimeSeries(values=vals['close'].values,
                            start_period=self._period_min, end_period=self._period_max,
                            kind=TimeSeriesKind.VALUES)
        currency_conversion_rate = self.__currency_conversion_rate(currency_to=self.currency.value)
        ts = ts * currency_conversion_rate
        return ts

    def __currency_conversion_rate(self, currency_to: Currency):
        currency_from = self.symbol.currency
        currency_rate = self.currency_symbols_registry \
            .convert(currency_from=currency_from,
                     currency_to=currency_to,
                     start_period=self._period_min,
                     end_period=self._period_max)
        currency_rate = TimeSeries(values=currency_rate['close'].values,
                                   start_period=currency_rate['period'].min(),
                                   end_period=currency_rate['period'].max(),
                                   kind=TimeSeriesKind.CURRENCY_RATE)
        return currency_rate

    def close(self):
        return copy.deepcopy(self.__values)

    def get_return(self, kind='values', real=False):
        if kind not in ['values', 'cumulative', 'ytd']:
            raise ValueError('`kind` is not in expected values')

        if kind == 'ytd':
            ror = self.get_return(kind='values', real=real)
            ror_ytd = ror.ytd()
            return ror_ytd

        ror = self.close().pct_change()

        if real:
            inflation = self.inflation(kind='values')
            ror = (ror + 1.) / (inflation + 1.) - 1.

        if kind == 'cumulative':
            ror = (ror + 1.).cumprod() - 1.

        return ror

    def risk(self, period='year'):
        """
        Returns risk of the asset

        :param period:
            month - returns monthly risk

            year - returns risk approximated to yearly value
        """
        p = Portfolio(assets=[self], weights=np.array([1.0]),
                      start_period=self._period_min,
                      end_period=self._period_max,
                      currency=self.currency.value)
        return p.risk(period=period)

    @contract(
        years_ago='int,>0|None|list[int,>0]',
        real='bool',
    )
    def get_cagr(self, years_ago=None, real=False):
        p = Portfolio(assets=[self], weights=np.array([1.0]),
                      start_period=self._period_min,
                      end_period=self._period_max,
                      currency=self.currency.value)
        return p.get_cagr(years_ago=years_ago, real=real)

    def inflation(self, kind: str, years_ago: int = None):
        ror = self.get_return()
        start_period = None if years_ago else ror.start_period
        return self.currency.inflation(kind=kind,
                                       start_period=start_period, end_period=ror.end_period,
                                       years_ago=years_ago)

    def __repr__(self):
        asset_repr = """\
            PortfolioAsset(
                 symbol: {},
                 currency: {},
                 period_min: {},
                 period_max: {}
            )""".format(self.symbol.identifier, self.currency,
                        self._period_min, self._period_max)
        return dedent(asset_repr)


@inject
class Portfolio:
    portfolio_currency_factory: PortfolioCurrencyFactory

    def __init__(self,
                 assets: List[PortfolioAsset],
                 weights: np.array,
                 start_period: pd.Period, end_period: pd.Period,
                 currency: Currency):
        self.weights = weights
        self.currency = self.portfolio_currency_factory.create(currency=currency)

        self._assets = [PortfolioAsset(symbol=a.symbol,
                                       start_period=start_period,
                                       end_period=end_period,
                                       currency=currency,
                                       portfolio=self,
                                       weight=w) for a, w in zip(assets, weights)]

    @property
    def assets(self) -> Dict[str, PortfolioAsset]:
        assets_dict = {a.symbol.identifier_str: a for a in self._assets}
        return assets_dict

    def risk(self, period='year'):
        """
        Returns risk of the asset

        :param period:
            month - returns monthly risk

            year - returns risk approximated to yearly value
        """
        if period == 'month':
            ror = self.get_return()
            return ror.std()
        elif period == 'year':
            ror = self.get_return()
            if ror.period_size < 12:
                raise Exception('year risk is requested for less than 12 months')

            mean = (1. + ror).mean()
            risk_monthly = self.risk(period='month')
            risk_yearly = ((risk_monthly ** 2 + mean ** 2) ** 12 - mean ** 24).sqrt()
            return risk_yearly
        else:
            raise Exception('unexpected value of `period` {}'.format(period))

    @contract(
        years_ago='int,>0|None|list[int,>0]',
        real='bool',
    )
    def get_cagr(self, years_ago=None, real=False):
        if years_ago is None:
            ror = self.get_return()
            years_total = ror.period_size / _MONTHS_PER_YEAR
            ror_c = (ror + 1.).prod()
            cagr = ror_c ** (1 / years_total) - 1.
            if real:
                inflation_cumulative = self.inflation(kind='cumulative')
                cagr = (cagr + 1.) / (inflation_cumulative + 1.) ** (1 / years_total) - 1.
            return cagr
        elif isinstance(years_ago, list):
            return np.array([self.get_cagr(years_ago=y, real=real)
                             for y in years_ago])
        elif isinstance(years_ago, int):
            ror = self.get_return()
            months_count = years_ago * _MONTHS_PER_YEAR
            if ror.period_size < months_count:
                return self.get_cagr(years_ago=None, real=real)

            ror_slice = self.get_return()[-months_count:]
            ror_slice_c = (ror_slice + 1.).prod()
            cagr = ror_slice_c ** (1 / years_ago) - 1.
            if real:
                inflation_cumulative = self.inflation(kind='cumulative',
                                                      years_ago=years_ago)
                cagr = (cagr + 1.) / (inflation_cumulative + 1.) ** (1 / years_ago) - 1.
            return cagr
        else:
            raise Exception('unexpected type of `years_ago`: {}'.format(years_ago))

    def get_return(self, kind='values', real=False) -> TimeSeries:
        if kind not in ['values', 'cumulative', 'ytd']:
            raise ValueError('`kind` is not in expected values')

        if kind == 'ytd':
            ror_assets = np.array([a.get_return(kind=kind, real=real) for a in self._assets])
            ror = (ror_assets * self.weights).sum()
            return ror

        ror_assets = np.array([a.get_return() for a in self._assets])
        ror = (ror_assets * self.weights).sum()

        if real:
            inflation = self.inflation(kind='values')
            ror = (ror + 1.) / (inflation + 1.) - 1.

        if kind == 'cumulative':
            ror = (ror + 1.).cumprod() - 1.

        return ror

    def inflation(self, kind: str, years_ago: int = None):
        ror = self.get_return()
        start_period = None if years_ago else ror.start_period
        return self.currency.inflation(kind=kind,
                                       start_period=start_period,
                                       end_period=ror.end_period,
                                       years_ago=years_ago)

    def __repr__(self):
        assets_repr = ', '.join(asset.symbol.identifier.__repr__() for asset in self._assets)
        portfolio_repr = """\
            Portfolio(
                 assets: {},
                 currency: {},
            )""".format(assets_repr, self.currency)
        return dedent(portfolio_repr)
